#
# Dataflow Analyis Engine Example
# Kevin Angstadt
# angstadt@virginia.edu
#

from copy import copy

# Intermediate code representation
# We support add, set, print, lables, and branch on equals
# Also, just variables and integers
class INSTR_Add(object):
    def __init__(self,dest,op1,op2):
        self.dest = dest
        self.op1 = op1
        self.op2 = op2
        
    def __str__(self):
        return '\t' + str(self.dest) + ' <- ' + str(self.op1) + ' + ' + str(self.op2)
    
class INSTR_Set(object):
    def __init__(self,dest,op1):
        self.dest = dest
        self.op1 = op1
        
    def __str__(self):
        return '\t' + str(self.dest) + ' <- ' + str(self.op1)
    
class INSTR_Print(object):
    def __init__(self,op1):
        self.op1 = op1
        
    def __str__(self):
        return '\tprint ' + str(self.op1)
    
class INSTR_Label(object):
    def __init__(self,label):
        self.label = label
        
    def __str__(self):
        return self.label + ':'
    
class INSTR_Beq(object):
    def __init__(self,op1,op2,label):
        self.op1 = op1
        self.op2 = op2
        self.label = label
    
    def __str__(self):
        return '\tbeq ' + str(self.op1) + ' = ' + str(self.op2) + ' -> ' + self.label

class OPERAND_Var(object):
    def __init__(self,name):
        self.name = name
    
    def __str__(self):
        return self.name

class OPERAND_Int(object):
    def __init__(self,val):
        self.val = val
    
    def __str__(self):
        return str(self.val)

# Control flow graph made of nodes (id,bb)
# and edges (id1,id2)
# Basic Blocks contain a list of instructions
class BasicBlock(object):
    def __init__(self,instructions):
        self.instructions = instructions
        
    def __str__(self):
        temp = ''
        for i in self.instructions:
            temp += str(i) + '\n'
        return temp
    
class CFGNode(object):
    def __init__(self,node_id,basic_block):
        self.id = node_id
        self.bb = basic_block
    
    def get_id(self):
        return self.id
    
    def __str__(self):
        return '\n// node ' + str(self.id) + '\n' + str(self.bb)
    
class CFGEdge(object):
    def __init__(self,source,sink):
        self.source = source
        self.sink = sink
        
    def __str__(self):
        return 'edge: ' + str(self.source) + ' -> ' + str(self.sink)
    
class CFG(object):
    def __init__(self):
        self.nodes = dict()
        self.edges = []
    
    def add_node(self,node):
        node_id = node.get_id()
        if node_id in self.nodes.keys():
            raise Exception('duplicate node id')
        
        self.nodes[node_id] = node
    
    def add_edge(self,edge):
        if edge in self.edges:
            raise Exception('duplicate edge')
        elif edge.source not in self.nodes.keys():
            raise Exception('source not found')
        elif edge.sink not in self.nodes.keys():
            raise Exception('sink not found')
        
        self.edges.append(edge)
        
    def __str__(self):
        temp = ''
        for key,n in self.nodes.iteritems():
            temp += str(n) + '\n'
        return temp

# abstract data-flow descriptor
# Needed:
#    * lattice of data-flow facts
#    * empty data-flow fact object
#    * transfer function
#    * join (lub) function
class AbstractDFDescriptor(object):
    # also need lattice objects!
    
    def not_yet_reached(self):
        raise Exception('Not implemented')
    
    def join(fact_list):
        raise Exception('Not implemented')
    
    def transfer(instruction, fact):
        raise Exception('Not implemented')

# data-flow analyis engine
# this uses a fixpoint-based approach to flow data through the CFG
# works with a given DFDescriptor object
class ForwardDataflow(object):
    # we store data-flow facts on edges of the CFG (we annotate the CFG edges)
    # another approach would be to add IN and OUT sets to each node
    class AnnotatedCFGEdge(CFGEdge):
        def __init__(self,edge,fact):
            CFGEdge.__init__(self,edge.source,edge.sink)
            self.fact = fact
        
        def set_fact(self,fact):
            self.fact = fact
        
        def get_fact(self):
            return self.fact
    
    class AnnotatedCFG(CFG):
        def __init__(self,df_descriptor):
            CFG.__init__(self)
            self.df = df_descriptor
            
        def __str__(self):
            temp = ''
            for _,n in self.nodes.iteritems():
                node_id = n.get_id()
                temp += '\n// node ' + str(node_id) + '\n'
                
                incoming_edges = [edge for edge in self.edges if edge.sink == node_id]
                incoming_facts = [edge.get_fact() for edge in incoming_edges]
                
                joined_fact = self.df.join(incoming_facts)
                
                temp += str(joined_fact) + '\n'
                
                for i in n.bb.instructions:
                    joined_fact = self.df.transfer(i,joined_fact)
                    temp += str(i) + '\n' + str(joined_fact) + '\n'
            return temp
                
    def __init__(self,cfg,df_descriptor):
        # convert the CFG to an AnnotatedCFG in preparation for performing
        # data-flow analysis
        self.df = df_descriptor
        self.cfg = self.AnnotatedCFG(self.df)
        for _,n in cfg.nodes.iteritems():
            self.cfg.add_node(n)
        for edge in cfg.edges:
            new_edge = self.AnnotatedCFGEdge(edge,self.df.not_yet_reached())
            self.cfg.add_edge(new_edge)
    
    # here is the heart of the algorithm
    def dataflow(self):
        fixpoint = False
        
        # repeatedly update the data-flow fact annotations until nothing more
        # can change.  Remember, a finite-height lattice and a monotonic
        # transfer function guaranteed to terminate
        while not fixpoint:
            # we'll set this back to false if we change anything
            fixpoint = True
            
            # iterate through all of the nodes in the CFG
            for node_id,n in self.cfg.nodes.iteritems():
                # given a node, find all facts that come into that node
                incoming_edges = [edge for edge in self.cfg.edges if edge.sink == node_id]
                incoming_facts = [edge.get_fact() for edge in incoming_edges]
                
                # take the join of these facts to find the best conservative
                # fact we can be certain of on entry
                joined_fact = self.df.join(incoming_facts)
                
                # transfer the data through the bb
                # this could be optimized by precomputing the transfer function
                # on the BB level, but left simple for clarity here
                for i in n.bb.instructions:
                    joined_fact = self.df.transfer(i,joined_fact)
                
                # joined_fact is now the outgoing fact for the BB
                # find all the outgoing edges
                outgoing_edges = [edge for edge in self.cfg.edges if edge.source == node_id]
                
                # assign the fact to the outgoing edge
                # if something changes, we have not yet reached a fixpoint
                for e in outgoing_edges:
                    if not (e.fact == joined_fact):
                        fixpoint = False
                        e.set_fact(joined_fact)

# Here is an example data-flow analysis descriptor to test out the framework
#
# We are only propagating a single data-flow fact through the CFG.  So, if we
# want to analyze many variables (in this case, several registers), we need
# to keep all of this inside of a single "fact".
#
# Here, we do that with a dictionary
class ConstantProp(AbstractDFDescriptor):
    # here we define our latice objects
    # the lattice has height 3, but infinite width
    
    # Top: not a constant
    class Top(object):
        def __str__(self):
            return 'Top'
        def __eq__(self,t):
            return isinstance(t,ConstantProp.Top)
    
    # Constant: exactly this constant
    class Constant(object):
        def __init__(self,val):
            self.val = val
        def __str__(self):
            return str(self.val)
        def __eq__(self,c):
            if not isinstance(c,ConstantProp.Constant):
                return False
            return self.val == c.val
        def add(self,c):
            return self.__class__(self.val + c.val)
    
    # Bottom: don't know
    class Bottom(object):
        def __str__(self):
            return 'Bottom'
        def __eq__(self,b):
            return isinstance(b,ConstantProp.Bottom)
    
    # our data-flow fact is a per-variable fact about every variable
    # for this case, it is a dictionary of 'varname' -> value bindings
    class DataFact(dict):
        def __str__(self):
            temp = ''
            for k in sorted(self):
                temp += str(k) + '=' + str(self[k]) + ' '
            return temp
        def __eq__(self,d):
            if(d.keys() != self.keys()):
                return False
            else:
                for k,_ in self.iteritems():
                    if not (self[k] == d[k]):
                        return False
            return True
                
    # default value for each annotated edge (empty, everything bottom)
    def not_yet_reached(self):
        return self.DataFact()
    
    # given an instruction and a data-flow fact, return a new fact
    def transfer(self,i,fact):
        # extract a variable from the fact
        def get_source(var,fact):
            if var in fact.keys():
                return fact[var]
            else:
                return self.Bottom()
        #make a copy the fact
        fact = copy(fact)
        
        raise Exception('not yet implemented')
        
        # The transfer function works on the instruction level of the program
        # (note that the composition of the transfer function for a basic block
        #  is handled by the dataflow engine)
        # 
        # Therefore, we need to focus on how data transfers through each
        # instruction.  We only have a couple instructions to worry about:
        #     - SET : update the destination register the value of the operand
        #     - ADD : dest <- op1 + op2
        #             get the information for op1 and op2
        #             if they are both constants, add them together
        #             for the value of dest
        #             otherwise, this is Top (we conservatively can't say anything)
        #     - BEQ, Label, Print : return the original fact (these don't affect
        #         data)
    
    # finally, we define how to join multiple facts
    # return a single fact
    def join(self,fact_list):
        raise Exception('not yet implemented')
    
        # you are given a list of facts, you need to join these into a single
        # fact.  Recall, we need to be conservative with this join.  We do this
        # by taking the least upper bound (lub) of the values for a given register
        # and returning that.
        # Ex: lub((nothing,Bottom),Constant(i)) = Constant(i)
        # Ex: lub(Constant(i),Constant(i)) = Top
        # Ex: lub(Constant(i),Top) = Top
        
    

def main():
    
    # A hand-crafted CFG, derived from this code: 
    # 
    # a = 5;
    # b = a + 2; 
    # if b == 2 {
    #      c = 7;
    #      d = 5;
    #      e = 9;
    # } else {
    #      c = 7;
    #      d = a;
    #      e = 8;
    # } 
    # print a;
    
    test_cfg_1 = CFG()

    test_cfg_1.add_node(
        CFGNode(1,
                BasicBlock([
                    INSTR_Set(OPERAND_Var('a'),OPERAND_Int(5)),
                    INSTR_Add(OPERAND_Var('b'),OPERAND_Var('a'),OPERAND_Int(2)),
                    INSTR_Beq(OPERAND_Var('b'),OPERAND_Int(2),'L_1')
                ])
            )
        )
    test_cfg_1.add_node(
        CFGNode(2,
                BasicBlock([
                    INSTR_Set(OPERAND_Var('c'),OPERAND_Int(7)),
                    INSTR_Set(OPERAND_Var('d'),OPERAND_Var('a')),
                    INSTR_Set(OPERAND_Var('e'),OPERAND_Int(8)),
                    INSTR_Beq(OPERAND_Int(0),OPERAND_Int(0),'L_2')
                ])
            )
        )
    test_cfg_1.add_node(
        CFGNode(3,
                BasicBlock([
                    INSTR_Label('L_1'),
                    INSTR_Set(OPERAND_Var('c'),OPERAND_Int(7)),
                    INSTR_Set(OPERAND_Var('d'),OPERAND_Int(5)),
                    INSTR_Set(OPERAND_Var('e'),OPERAND_Int(9))
                ])
            )
        )
    test_cfg_1.add_node(
        CFGNode(4,
                BasicBlock([
                    INSTR_Label('L_2'),
                    INSTR_Print(OPERAND_Var('a'))
                ])
            )
        )
    
    test_cfg_1.add_edge(CFGEdge(1,2))
    test_cfg_1.add_edge(CFGEdge(1,3))
    test_cfg_1.add_edge(CFGEdge(2,4))
    test_cfg_1.add_edge(CFGEdge(3,4))
    
    print test_cfg_1
    
    # make the data-flow analyzer!
    df = ForwardDataflow(test_cfg_1,ConstantProp())
    
    # do the data-flow!
    df.dataflow()
    
    print df.cfg
    
if __name__ == '__main__':
    main()